import matplotlib.pyplot as plt
import statistics
import sys
import os
import matplotlib
import numpy as np
matplotlib.rcParams['text.usetex'] = True
matplotlib.rcParams['text.latex.preamble'] = '\\usepackage{sfmath}'

FONTSZ=18
SMALL=False
COLORS = [
    [],
    ['#fec44f'],
    ['#fff7bc', '#d95f0e'],
    ['#fff7bc', '#fec44f', '#d95f0e'],
    ['#ffffd4', '#fed98e', '#fe9929', '#cc4c02'],
    ['#ffffd4', '#fed98e', '#fe9929', '#cc4c02'] * 2,
    ['#ffffd4', '#fed98e', '#fe9929', '#cc4c02'] * 2,
    ['#ffffd4', '#fed98e', '#fe9929', '#cc4c02'] * 2,
    ['#ffffd4', '#fed98e', '#fe9929', '#cc4c02'] * 2,
]

PROGRAM = [
    "xml-tu",
#   'xml',
#    "xml-custom-hashmap",
    'HAMT-1',
    'HAMT-10',
#    'HAMTS-1',
#    'HAMTS-10',
    'CHAMP-1',
    'CHAMP-10',
    'avrora',
    'fop',
    'h2',
    'jython',
    'luindex',
    'lusearch',
#    'lusearch-fix',
    'pmd',
    'sunflow',
    'xalan',
]

PROGRAM_NAME = {
    "xml-tu": "XML",
}

def program_name(p):
    if p in PROGRAM_NAME:
        return PROGRAM_NAME[p]
    else:
        return p

def config_string(c):
    return "%(setting)s-%(vm)s-%(analysis)s" % c

def load_data(file_name):
    data = {}
    with open(file_name) as f:
        head = f.readline().strip()
        confs = head.split(",")[1:]
        for conf in confs:
            data[conf] = {}
        for line in f.readlines():
            fields = line.strip().split(",")
            key = fields[0]
            for i, field in enumerate(fields[1:]):
                if field != "":
                    data[confs[i]][key] = float(field)
    return data


def load_all_data():
    data = {}
    for b in PROGRAM:
        data[b] = load_data("csv/count_%s.csv" % b)
    return data

def load_time(dir_name, configs):
    def load_time1(b, c):
        p = {
            "s": c["setting"],
            "v": c["vm"],
            "a": c["analysis"],
            "b": b
        }
        file_name= dir_name + "/%(s)s_%(b)s_%(v)s_%(a)s.txt" % p
        numbers = []
        try:
            with open(file_name) as f:
                for line in f.readlines():
                    numbers.append(float(line))
            if SMALL:
                return np.mean(numbers[1:])
            else:
                return np.mean(numbers[5:])
        except:
            return 0
    
    data = {}
    for b in PROGRAM:
        data[b] = {}
        for c in configs:
            data[b][config_string(c)] = load_time1(b, c)

    return data

def load_all_time():
    configs = [
        {"setting": "all",  "vm": "cw", "analysis": "none"},
        {"setting": "none", "vm": "original", "analysis": "none"}
    ]
    time = load_time("results", configs)
    return time

def make_numbers(count, time, configs):
    data = []
    for conf in configs:
        name = config_string(conf)
        dt = []
        for b in PROGRAM:
            t = time[b]["all-cw-none"]
#            t = time[b]["none-original-none"]            
            print(b, name, t)
            if "handshake" not in count[b][name]:
                dt.append(None)
            else:
                dt.append(count[b][name]['handshake'] / t)
        data.append(dt)
    return data

def draw(data, labels, hlspan, outfile):
    def draw1(ax, dt, label, index):
        width = 1.0/(len(data) + 1)
        left = [x + width * index for x in range(len(dt))]
        height = dt
        csz = 7.0 / (len(data) + 1)
        ax.bar(left, height, color=COLORS[len(labels)][index],
               edgecolor="black",
               capsize=csz, label=label, width=width)

    def add_xticks(ax):
        ax.set_xticks([x + 0.5 for x in range(len(PROGRAM))],
                      [program_name(x) for x in PROGRAM],
                      rotation=60, ha="right")
        ax.tick_params(bottom=False)


    fig = plt.figure()
    ax = fig.add_subplot(111)

    ymax = max([max(dt) for dt in data]) * 1.4
    xmin = -1
    xmax = len(PROGRAM) + 1
    ax.set_ylim(0, ymax)
    ax.set_xlim(xmin, xmax)
    y = hlspan
    while y < ymax:
        plt.hlines(y, xmin, xmax,
                   lw=0.5, color='black', zorder=1)
        y += hlspan
    
    for i, dt in enumerate(data):
        draw1(ax, dt, labels[i], i)
    add_xticks(ax)
    ax.set_ylabel("handshakes ($\\times 10^3$/sec)", fontsize=FONTSZ)
    ax.tick_params(labelsize=FONTSZ)
    plt.legend(fontsize=FONTSZ)
    plt.tight_layout()
    if outfile:
        plt.savefig(outfile)
    else:
        plt.show()

def main():
    count = load_all_data()
    time = load_all_time()

    print(time)

    configs = [
        {"setting": "all",  "vm": "cw", "analysis": "none"},
        {"setting": "all", "vm": "cw", "analysis": "h"}
    ]
    labels = [
        "no analysis",
        "with static analysis"
    ]
    
    data = make_numbers(count, time, configs)
    for i in range(len(data[0])):
        valid = True
        for dt in data:
            if dt[i] == None:
                valid = False
        if not valid:
            PROGRAM_NAME[PROGRAM[i]] = ""
            for dt in data:
                dt[i] = 0
    draw(data, labels, 20, "handshake.pdf")

if len(sys.argv) >= 2 and sys.argv[1] == "-small":
    SMALL=True
   
main()
