import matplotlib.pyplot as plt
import statistics
import sys
import os
import matplotlib
from scipy.stats import gmean
matplotlib.rcParams['text.usetex'] = True
matplotlib.rcParams['text.latex.preamble'] = '\\usepackage{sfmath}'

FONTSZ=18
SMALL=False

COLORS = [
    [],
    ['#fec44f'],
    ['#fff7bc', '#d95f0e'],
    ['#fff7bc', '#fec44f', '#d95f0e'],
    ['#ffffd4', '#fed98e', '#fe9929', '#cc4c02'],
    ['#ffffd4', '#fed98e', '#fe9929', '#cc4c02'] * 2,
    ['#ffffd4', '#fed98e', '#fe9929', '#cc4c02'] * 2,
    ['#ffffd4', '#fed98e', '#fe9929', '#cc4c02'] * 2,
    ['#ffffd4', '#fed98e', '#fe9929', '#cc4c02'] * 2,
]

PROGRAM = [
    "xml-tu",
#   'xml',
#    "xml-custom-hashmap",
    'HAMT-1',
    'HAMT-10',
#    'HAMTS-1',
#    'HAMTS-10',
    'CHAMP-1',
    'CHAMP-10',
    'avrora',
    'fop',
    'h2',
    'jython',
    'luindex',
    'lusearch',
#    'lusearch-fix',
    'pmd',
    'sunflow',
    'xalan',
]

PROGRAM_NAME = {
    "xml-tu": "XML",
}

def program_name(p):
    if p in PROGRAM_NAME:
        return PROGRAM_NAME[p]
    else:
        return p

BENCHMARK_PARAM = {
    'sxml' : {'warmup': 1, 'actual': 2},
}

def config_to_name(benchmark, conf):
    return ("%(setting)s_"+benchmark+"_%(vm)s_%(analysis)s.txt") % conf

def load_data(dir_name, confs, warmup = 5, min_actual = 10):
    if SMALL:
        warmup = 1
        min_actual = 1

    def load_numbers(file_name, benchmark):
        nw = warmup
        na = min_actual
        if benchmark in BENCHMARK_PARAM:
            p = BENCHMARK_PARAM[benchmark]
            if 'warmup' in p:
                nw = p['warmup']
            if 'actual' in p:
                na = p['actual']
        if os.path.isfile(file_name):
            with open(file_name) as f:
                ns = [float(x) for x in f.readlines()]
                if len(ns) < nw + na:
                    raise Exception("too few data in " + file_name)
                return ns[nw:]
        else:
            print("no data for "+file_name)
        return None

    def load_for_conf(conf):
        data = []
        for b in PROGRAM:
            name = dir_name + '/' + config_to_name(b, conf)
            data.append(load_numbers(name, b))
        return data

    return [load_for_conf(x) for x in confs]

def convert_to_avg_err(data):
    def convert1(numbers):
        if numbers:
            avg = statistics.mean(numbers)
            sd = statistics.stdev(numbers)
            return (avg, sd, sd)
        else:
            return (0, 0, 0)

    return [[convert1(y) for y in each_conf] for each_conf in data]


def normalize(baseline, data):
    def normalize_for_conf(data_for_conf):
        normalized = []
        for i, (time, _, _) in enumerate(baseline):
            if time == 0:
                print("warnnig: baseline has 0")
                normalized.append((-1, 0, 0))
            else:
                normalized.append(tuple(x / time for x in data_for_conf[i]))
        return normalized

    return [normalize_for_conf(x) for x in data]

def draw(data, labels, hlspan, outfile, show_geomean):
    def draw1(ax, dt, label, index):
        width = 1.0/(len(data) + 1)
        left = [x + width * index for x in range(len(dt))]
        height = [x[0] for x in dt]
        yerr = ([x[1] * 2 for x in dt],
                [x[2] * 2 for x in dt])
        if show_geomean:
            left.append(len(dt) + width * index)
            gm = gmean(height)
            print(label + " gmean = " +str(gm))
            height.append(gm)
            yerr[0].append(0)
            yerr[1].append(0)
        csz = 7.0 / (len(data) + 1)
        ax.bar(left, height, yerr=yerr, color=COLORS[len(labels)][index],
               edgecolor="black",
               capsize=csz, label=label, width=width)

    def add_xticks(ax):
        positions = [x + 0.5 for x in range(len(PROGRAM))]
        labels = [program_name(x) for x in PROGRAM]
        if show_geomean:
            positions.append(len(PROGRAM) + 0.5)
            labels.append("geo.mean")
        ax.set_xticks(positions, labels, rotation=60, ha="right")
        ax.tick_params(bottom=False)


    fig = plt.figure()
    ax = fig.add_subplot(111)

    ymax = max([max(x[0] for x in dt) for dt in data]) * 1.2
    xmin = -1
    xmax = len(PROGRAM) + 1
    ax.set_ylim(0, ymax)
    ax.set_xlim(xmin, xmax)
    y = hlspan
    while y < ymax:
        plt.hlines(y, xmin, xmax,
                   lw=0.5, color='black', zorder=1)
        y += hlspan
    
    for i, dt in enumerate(data):
        draw1(ax, dt, labels[i], i)
    add_xticks(ax)
    ax.set_ylabel("normalized execution time", fontsize=FONTSZ)
    ax.tick_params(labelsize=FONTSZ)
    plt.legend(fontsize=FONTSZ)
    plt.tight_layout()
    if outfile:
        plt.savefig(outfile)
    else:
        plt.show()

def draw_bd(data, data2, labels, hlspan, outfile, show_geomean):
    def draw1(ax, dt, label, index, main_bar, geo_mean):
        width = 1.0/(len(data) + 1)
        csz = 7.0 / (len(data) + 1)
        if geo_mean:
            print(exclude)
            left = [len(dt) + width * index]
            gm = gmean([x[0] for i,x in enumerate(dt) if i not in exclude])
            print(label + " gmean = "+str(gm))
            height = [gm]
        else:
            left = [x + width * index for x in range(len(dt))]
            height = [x[0] for x in dt]

        if main_bar:
            color = 'white'
            if geo_mean:
                ax.bar(left, height, color=color,
                       edgecolor="black",
                       capsize=csz, width=width)
            else:
                yerr = ([x[1] * 2 for x in dt],
                        [x[2] * 2 for x in dt])
                ax.bar(left, height, yerr=yerr, color=color,
                       edgecolor="black",
                       capsize=csz, width=width)
        else:
            color = COLORS[len(labels)][index]
            if geo_mean:
                ax.bar(left, height, color=color,
                       edgecolor="black",
                       capsize=csz, width=width)
            else:
                ax.bar(left, height, color=color,
                       edgecolor="black",
                       capsize=csz, label=label, width=width)

    def add_xticks(ax):
        positions = [x + 0.5 for x in range(len(PROGRAM))]
        labels = [program_name(x) for x in PROGRAM]
        if show_geomean:
            positions.append(len(PROGRAM) + 0.5)
            labels.append("geo.mean")
        ax.set_xticks(positions, labels, rotation=60, ha="right")
        ax.tick_params(bottom=False)

    exclude = []
    for dt in data:
        for i, x in enumerate(dt):
            if x[0] == 0:
                exclude.append(i)
        
    fig = plt.figure()
    ax = fig.add_subplot(111)

    ymax = max([max(x[0] for x in dt) for dt in data]) * 1.4
    xmin = -1
    xmax = len(PROGRAM) + 1
    ax.set_ylim(0, ymax)
    ax.set_xlim(xmin, xmax)
    y = hlspan
    while y < ymax:
        plt.hlines(y, xmin, xmax,
                   lw=0.5, color='black', zorder=1)
        y += hlspan
    
    for i, dt in enumerate(data):
        draw1(ax, dt, labels[i], i, True, False)
    for i, dt in enumerate(data2):
        draw1(ax, dt, labels[i], i, False, False)
    if show_geomean:
        for i, dt in enumerate(data):
            draw1(ax, dt, labels[i], i, True, True)
        for i, dt in enumerate(data2):
            draw1(ax, dt, labels[i], i, False, True)
    add_xticks(ax)
    ax.set_ylabel("normalized execution time", fontsize=FONTSZ)
    ax.tick_params(labelsize=FONTSZ)
    plt.legend(fontsize=FONTSZ)
    plt.tight_layout()
    if outfile:
        plt.savefig(outfile)
    else:
        plt.show()

        
def fig_none_durable(baseline, dir_name, outfile=None):
    configs = [
        {"setting": "none", "vm": "ww", "analysis": "none"},
        {"setting": "none", "vm": "cw", "analysis": "none"},
    ]
    labels = [
        "writer-wait approach",
        "copier-wait approach"
    ]
    data = load_data(dir_name, configs)
    data = convert_to_avg_err(data)
    data = normalize(baseline, data)
    draw(data, labels, 0.25, outfile, True)

def fig_none_durable_bd(baseline, dir_name, outfile=None):
    configs = [
        {"setting": "none", "vm": "ww", "analysis": "none"},
        {"setting": "none", "vm": "cw", "analysis": "none"},
    ]
    configs_nomfence = [
        {"setting": "none", "vm": "wwx", "analysis": "none"},
        {"setting": "none", "vm": "cw", "analysis": "none"},
    ]
    labels = [
        "writer-wait approach",
        "copier-wait approach"
    ]
    data = load_data(dir_name, configs)
    data = convert_to_avg_err(data)
    data = normalize(baseline, data)
    data_nf = load_data(dir_name, configs_nomfence)
    data_nf = convert_to_avg_err(data_nf)
    data_nf = normalize(baseline, data_nf)
    
    draw_bd(data, data_nf, labels, 0.25, outfile, True)
    
def fig_all_durable(baseline, dir_name, outfile=None):
    configs = [
        {"setting": "all", "vm": "cw", "analysis": "none"},
        {"setting": "all", "vm": "cw", "analysis": "h"},
#        {"setting": "all", "vm": "cw", "analysis": "bh"},
    ]
    labels = [
        "copier-wait (no analysis)",
        "copier-wait (handshake elision)",
        "copier-wait (both elision)",
    ]
    data = load_data(dir_name, configs)
    data = convert_to_avg_err(data)
    data = normalize(baseline, data)
    draw(data, labels, 0.5, outfile)

def fig_all_durable_bd(baseline, dir_name, outfile=None):
    configs = [
        {"setting": "all", "vm": "cw", "analysis": "none"},
#        {"setting": "all", "vm": "cw", "analysis": "b"},
        {"setting": "all", "vm": "cw", "analysis": "h"},
#        {"setting": "all", "vm": "cwy", "analysis": "h"},
    ]
    configs_nohandshake = [
        {"setting": "all", "vm": "cwx", "analysis": "none"},
#        {"setting": "all", "vm": "cwx", "analysis": "b"},
        {"setting": "all", "vm": "cwx", "analysis": "h"},
#        {"setting": "all", "vm": "cwy", "analysis": "h"},
    ]
    labels = [
        "no analysis",
#        "copier-wait (barrier elision)",
        "with static analysis",
#        "copier-wait (hs elision+redundant verify)",
    ]
    data = load_data(dir_name, configs)
    data = convert_to_avg_err(data)
    data = normalize(baseline, data)
    data_nohs = load_data(dir_name, configs_nohandshake)
    data_nohs = convert_to_avg_err(data_nohs)
    data_nohs = normalize(baseline, data_nohs)
    draw_bd(data, data_nohs, labels, 0.5, outfile, True)
    
def print_data(baselines, dir_name):
    def do_print(label, dt):
        print(label)
        for (x,_,_) in dt:
            print(x)
        print()
    
    def print_results(baseline, configs):
        data = load_data(dir_name, configs)
        data = convert_to_avg_err(data)
        data = normalize(baseline, data)
        for i, dt in enumerate(data):
            label = "%(setting)s-%(vm)s-%(analysis)s" % configs[i]
            do_print(label, dt)

        
    configs = [
        {"setting": "none", "vm": "original", "analysis": "none"},
        {"setting": "none", "vm": "ww", "analysis": "none"},
        {"setting": "none", "vm": "cw", "analysis": "none"},
    ]
    print_results(baselines[0], configs)

    configs = [
        {"setting": "all", "vm": "ww", "analysis": "none"},
        {"setting": "all", "vm": "cw", "analysis": "none"},
        {"setting": "all", "vm": "cw", "analysis": "h"},
#        {"setting": "all", "vm": "cw", "analysis": "bh"},
    ]
    print_results(baselines[1], configs)

def csv_print(dir_name, configs, filename, baseline):
    data = load_data(dir_name, configs)
    data = convert_to_avg_err(data)
    if baseline:
        data = normalize(baseline, data)

    with open(filename, "w") as out:
        labels = ['benchmark']
        for i, _ in enumerate(data):
            label = "%(setting)s-%(vm)s-%(analysis)s" % configs[i]
            labels.append(label)
        out.write(','.join(labels) + '\n')
        
        for i, b in enumerate(PROGRAM):
            vals = [b]
            for dt in data:
                vals.append(str(dt[i][0]))
            out.write(','.join(vals) + '\n')

def csv_raw(dir_name):
    configs = [
        {"setting": "none", "vm": "original", "analysis": "none"},
        {"setting": "none", "vm": "ww", "analysis": "none"},
        {"setting": "none", "vm": "cw", "analysis": "none"},
        {"setting": "all", "vm": "ww", "analysis": "none"},
        {"setting": "all", "vm": "cw", "analysis": "none"},
        {"setting": "all", "vm": "cw", "analysis": "h"},
#        {"setting": "all", "vm": "cw", "analysis": "b"},
#        {"setting": "all", "vm": "cw", "analysis": "bh"},
        {"setting": "all", "vm": "cwx", "analysis": "none"},
        {"setting": "all", "vm": "cwx", "analysis": "h"},
#        {"setting": "all", "vm": "cwx", "analysis": "b"},
#        {"setting": "all", "vm": "cwx", "analysis": "bh"}
    ]
    csv_print(dir_name, configs, "csv/raw.csv", None)

def csv_none_durable(dir_name, baseline):
    configs = [
        {"setting": "none", "vm": "original", "analysis": "none"},
        {"setting": "none", "vm": "ww", "analysis": "none"},
        {"setting": "none", "vm": "cw", "analysis": "none"},
    ]
    csv_print(dir_name, configs, "csv/none_durable.csv", baseline)

def csv_all_durable(dir_name, baseline):
    configs = [
        {"setting": "all", "vm": "ww", "analysis": "none"},
        {"setting": "all", "vm": "cw", "analysis": "none"},
        {"setting": "all", "vm": "cw", "analysis": "h"},
#        {"setting": "all", "vm": "cw", "analysis": "b"},
#        {"setting": "all", "vm": "cw", "analysis": "bh"},
        {"setting": "all", "vm": "cwx", "analysis": "none"},
        {"setting": "all", "vm": "cwx", "analysis": "h"},
#        {"setting": "all", "vm": "cwx", "analysis": "b"},
#        {"setting": "all", "vm": "cwx", "analysis": "bh"}
    ]
    csv_print(dir_name, configs, "csv/all_durable.csv", baseline)
    

def main():
    dir_name = "results"
    
    c = [
        {"setting": "none", "vm": "ww", "analysis": "none"},
        {"setting": "all", "vm": "ww", "analysis": "none"},
        {"setting": "none", "vm": "original", "analysis": "none"}
    ]
    baselines = convert_to_avg_err(load_data(dir_name, c))

#    print_data(baselines, dir_name)
    fig_none_durable(baselines[2], dir_name, 'none_durable.pdf')
    fig_all_durable_bd(baselines[1], dir_name, 'all_durable.pdf')
    csv_raw(dir_name)
    csv_none_durable(dir_name, baselines[0])
    csv_all_durable(dir_name, baselines[1])


if len(sys.argv) >= 2 and sys.argv[1] == "-small":
    SMALL=True

main()
               
